# -*- coding: utf-8 -*-
# Copyright (c) 2012 by Pablo Martín <goinnn@gmail.com>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

import csv
import sys

PY3 = sys.version > '3'

try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO
import codecs

import django_tables2 as tables

from django.utils.translation import ugettext as _
from django.http import HttpResponse
from django.utils.html import strip_tags

from django_tables2_reports.csv_to_excel import EXCEL_SUPPORT, convert_to_excel
from django_tables2_reports.utils import DEFAULT_PARAM_PREFIX, generate_prefixto_report


# Unicode CSV writer, copied direct from Python docs:
# http://docs.python.org/2/library/csv.html

class UnicodeWriter:
    """
    A CSV writer which will write rows to CSV file "f",
    which is encoded in the given encoding.
    """

    def __init__(self, f, dialect=csv.excel, encoding="utf-8", **kwds):
        # Redirect output to a queue
        self.queue = StringIO()
        self.writer = csv.writer(self.queue, dialect=dialect, **kwds)
        self.stream = f
        self.encoder = codecs.getincrementalencoder(encoding)()

    def writerow(self, row):
        if PY3:
            self.writer.writerow([s for s in row])
            # Fetch UTF-8 output from the queue ...
            data = self.queue.getvalue()
        else:
            self.writer.writerow([s.encode("utf-8") for s in row])
            # Fetch UTF-8 output from the queue ...
            data = self.queue.getvalue()
            data = data.decode("utf-8")
        # ... and reencode it into the target encoding
        data = self.encoder.encode(data)
        # write to the target stream
        self.stream.write(data)
        # empty queue
        self.queue.truncate(0)

    def writerows(self, rows):
        for row in rows:
            self.writerow(row)


class TableReport(tables.Table):

    def __init__(self, *args, **kwargs):
        if not 'template' in kwargs:
            kwargs['template'] = 'django_tables2_reports/table.html'
        prefix_param_report = kwargs.pop('prefix_param_report', DEFAULT_PARAM_PREFIX)
        super(TableReport, self).__init__(*args, **kwargs)
        self.param_report = generate_prefixto_report(self, prefix_param_report)
        self.formats = [(_('CSV Report'), 'csv')]
        if EXCEL_SUPPORT:
            self.formats.append((_('XLS Report'), 'xls'))

    def as_report(self, request, format='csv'):
        if format == 'csv':
            return self.as_csv(request)
        elif format == 'xls':
            return self.as_xls(request)
        raise ValueError("This format %s is not accepted" % format)

    def as_csv(self, request):
        response = HttpResponse()
        csv_writer = UnicodeWriter(response)

        csv_header = [column.header for column in self.columns]
        csv_writer.writerow(csv_header)

        for row in self.rows:
            csv_writer.writerow([strip_tags(cell) for column, cell in row.items()])

        return response

    def as_xls(self, request):
        return self.as_csv(request)

    def treatement_to_response(self, response, format='csv'):
        if format == 'xls':
            convert_to_excel(response)
        return response



from django.test import TestCase
from django.http import HttpRequest


import django_tables2
import django_tables2_reports.tables
import django_tables2_reports.views


class TableReportForTesting(django_tables2_reports.tables.TableReport):
    name = django_tables2.Column()
    item_num = django_tables2.Column()


class ReportTableViewForTesting(django_tables2_reports.views.ReportTableView):
    table_class = TableReportForTesting


class TestCsvGeneration(TestCase):
    """Test csv generation on sample table data."""

    def test_csv_simple_input(self):
        """Test ability to generate csv with simple input data."""

        # Mix of integer and string data.  Ensure that commas and
        # quotes are escaped properly.
        data = [
            {
                'name': 'Normal string',
                'item_num': 1,
            },
            {
                'name': 'String, with, commas',
                'item_num': 2,
            },
            {
                'name': 'String with " quote',
                'item_num': 3,
            },
        ]

        table = TableReportForTesting(data)
        response = table.as_csv(HttpRequest())

        # Expect cells containing commas to be escaped with quotes.
        self.assertEqual(
            response.content,
            'Name,Item Num\r\n'
            'Normal string,1\r\n'
            '"String, with, commas",2\r\n'
            '"String with "" quote",3\r\n')

    def test_csv_with_unicode(self):
        """Test that unicode cell values are converted correctly to csv."""

        data = [
            {
                'name': 'Normal string',
                'item_num': 1,
            },
            {
                'name': u'String with ' + unichr(0x16c) + ' char',
                'item_num': 2,
            },
        ]

        table = TableReportForTesting(data)
        response = table.as_csv(HttpRequest())

        # Expect csv content to be utf-8 encoded.
        self.assertEqual(
            response.content,
            ('Name,Item Num\r\n'
             'Normal string,1\r\n'
             'String with ' + unichr(0x16c) + ' char,2\r\n').encode('utf-8'))


    def test_csv_no_pagination(self):
        """Ensure that table pagination doesn't affect output."""

        data = [
            {
                'name': 'page 1',
                'item_num': 1,
            },
            {
                'name': 'page 2',
                'item_num': 2,
            },
        ]

        table = TableReportForTesting(data)
        table.paginate(per_page=1)

        response = table.as_csv(HttpRequest())

        # Ensure that even if table paginated, output is all row
        # data.
        self.assertEqual(
            response.content,
            ('Name,Item Num\r\n'
             'page 1,1\r\n'
             'page 2,2\r\n')
        )


class TestExcelGeneration(TestCase):

    def test_excel_simple_input(self):
        """Test ability to generate excel output with simple input data."""

        if not django_tables2_reports.csv_to_excel.EXCEL_SUPPORT:
            return

        # Mix of integer and string data.  Ensure that commas and
        # quotes are escaped properly.
        data = [
            {
                'name': 'Normal string',
                'item_num': 1,
            },
            {
                'name': 'String, with, commas',
                'item_num': 2,
            },
            {
                'name': 'String with " quote',
                'item_num': 3,
            },
            {
                'name': u'String with ' + unichr(0x16c) + ' char',
                'item_num': 4,
            }
        ]

        table = TableReportForTesting(data)
        response = table.treatement_to_response(
            table.as_csv(HttpRequest()),
            format='xls')

        # No assertions.  Expect conversion to xls to succeed, even with
        # unicode chars.  Uncomment the following line and open test-file.xls
        # manually using Excel to verify that content is correct.
        open('test-file.xls', 'wb').write(response.content)



 # Copyright (c) 2010 by Yaco Sistemas <pmartin@yaco.es>
 #
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU Lesser General Public License as published by
 # the Free Software Foundation, either version 3 of the License, or
 # (at your option) any later version.
 #
 # This program is distributed in the hope that it will be useful,
 # but WITHOUT ANY WARRANTY; without even the implied warranty of
 # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 # GNU Lesser General Public License for more details.
 #
 # You should have received a copy of the GNU Lesser General Public License
 # along with this programe.  If not, see <http://www.gnu.org/licenses/>.

# Based on http://sujitpal.blogspot.com/2007/02/python-script-to-convert-csv-files-to.html
# Get to https://github.com/Yaco-Sistemas/django-autoreports/blob/master/autoreports/csv_to_excel.py

import csv
import sys

PY3 = sys.version > '3'

try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO

import collections

# Autodetect library to use for xls writing.  Default to xlwt.
EXCEL_SUPPORT = None
if PY3:
    try:
        import xlwt
        EXCEL_SUPPORT = 'xlwt'
    except ImportError:
        pass
else:
    try:
        import xlwt
        EXCEL_SUPPORT = 'xlwt'
    except ImportError:
        pass

    if EXCEL_SUPPORT is None:
        try:
            import pyExcelerator
            EXCEL_SUPPORT = 'pyexcelerator'
        except ImportError:
                pass


def openExcelSheet():
    """ Opens a reference to an Excel WorkBook and Worksheet objects """
    workbook = pyExcelerator.Workbook()
    worksheet = workbook.add_sheet("Sheet 1")
    return workbook, worksheet


def validateOpts(response):
    """ Returns option values specified, or the default if none """
    titlePresent = False
    linesPerFile = -1
    sepChar = ","
    return titlePresent, linesPerFile, sepChar


def writeExcelHeader(worksheet, titleCols):
    """ Write the header line into the worksheet """
    cno = 0
    for titleCol in titleCols:
        worksheet.write(0, cno, titleCol)
        cno = cno + 1


def writeExcelRow(worksheet, lno, columns):
    """ Write a non-header row into the worksheet """
    cno = 0
    for column in columns:
        worksheet.write(lno, cno, column.decode('utf-8'))
        cno = cno + 1


def closeExcelSheet(response, workbook):
    """ Saves the in-memory WorkBook object into the specified file """
    response.content = workbook.get_biff_data()


def convert_to_excel_pyexcelerator(response):
    titlePresent, linesPerFile, sepChar = validateOpts(response)
    workbook, worksheet = openExcelSheet()
    fno = 0
    lno = 0
    titleCols = []
    content = StringIO(response.content)
    reader = csv.reader(content)
    for line in reader:
        if (lno == 0 and titlePresent):
            if (len(titleCols) == 0):
                titleCols = line
            writeExcelHeader(worksheet, titleCols)
        else:
            writeExcelRow(worksheet, lno, line)
        lno = lno + 1
        if (linesPerFile != -1 and lno >= linesPerFile):
            fno = fno + 1
            lno = 0
    closeExcelSheet(response, workbook)


# A reasonable approximation for column width is based off zero character in
# default font.  Without knowing exact font details it's impossible to
# determine exact auto width.
# http://stackoverflow.com/questions/3154270/python-xlwt-adjusting-column-widths?lq=1
def get_xls_col_width(text, style):
    return int((1 + len(text)) * 256)


def write_xlwt_row(ws, lno, cell_text, cell_widths, style=None):
    """Write row of utf-8 encoded data to worksheet, keeping track of maximum
    column width for each cell.
    """
    if style is None:
        style = xlwt.Style.default_style

    for cno, utf8_text in enumerate(cell_text):
        cell_text = utf8_text
        if not PY3:
            cell_text = cell_text.decode('utf-8')
        ws.write(lno, cno, cell_text, style)
        cell_widths[cno] = max(cell_widths[cno],
                               get_xls_col_width(cell_text, style))


def convert_to_excel_xlwt(response):
    """Replace HttpResponse csv content with excel formatted data using xlwt
    library.
    """
    # Styles used in the spreadsheet.  Headings are bold.
    header_font = xlwt.Font()
    header_font.bold = True

    header_style = xlwt.XFStyle()
    header_style.font = header_font

    wb = xlwt.Workbook()
    ws = wb.add_sheet('Sheet 1')

    # Cell width information kept for every column, indexed by column number.
    cell_widths = collections.defaultdict(lambda: 0)
    if PY3:
        content = StringIO(response.content.decode('utf-8').replace('\x00', ''))
    else:
        content = StringIO(response.content)
    reader = csv.reader(content)
    for lno, line in enumerate(reader):
        if lno == 0:
            style = header_style
        else:
            style = None

        write_xlwt_row(ws, lno, line, cell_widths, style)

    # Roughly autosize output column widths based on maximum column size.
    for col, width in cell_widths.items():
        ws.col(col).width = width

    response.content = ''
    wb.save(response)


def convert_to_excel(response):
    if EXCEL_SUPPORT == 'xlwt':
        convert_to_excel_xlwt(response)
    elif EXCEL_SUPPORT == 'pyexcelerator':
        convert_to_excel_pyexcelerator(response)
    else:
        raise RuntimeError("No support for xls generation available")
